using Godot;
using System;

public class BossClawDemon : KinematicBody2D
{
    // nodes
    Node stats;
    Node playerVariables;
    RayCast2D longRayLeft;
    RayCast2D longRayRight;
    RayCast2D shortRayLeft;
    RayCast2D shortRayRight;
    Area2D areaStopRunAttack;

    AnimationPlayer animationPlayer;
    AnimationPlayer hitAnimationPlayer;
    Timer introTimer;
    Timer idleTimer;
    Timer battlecryTimer;
    Timer dashTimer;
    Timer jumpPrepareTimer;
    Timer openGateTimer;
    Timer rockSpawnTimer;
    Area2D playerDetection;
    KinematicBody2D playerBody;
    Camera2D camera2D;
    Area2D swordHitbox;
    CollisionShape2D hurtboxCollision;
    CollisionShape2D hitboxCollision;
    PackedScene dustParticles;
    PackedScene fallingRock;

    //UI
    ProgressBar healthBar;
    Label bossNameLabel;

    PackedScene audioStreamPlayer;
    AudioStreamSample asHit;
    AudioStreamSample asDash;
    AudioStreamSample asJump;
    AudioStreamSample asImpact;
    AudioStreamSample asDeath;

    [Signal] public delegate void bossKilledOpenTheGateSignal();

    Random r = new Random();

    // constants
    const int ACCELERATION = 1200;
    const int MAX_SPEED = 150;
    const int JUMP_FORCE = 350;
    const int MAX_FALL_SPEED = 280;
    const float AIR_TIME = 1.15f;
    const int MAX_DASH_SPEED = 350;
    const int GRAVITY = 600;
    const float FRICTION = 0.3f;
    const int MAX_HEALTH = 60;
    const int SOUL = 100;
    // timer and shake constants
    const float SHAKE_STRENGTH = 1;
    const int BATTLECRY_TIME = 3;
    const int ROAR_ATTACK_TIME = 10;
    // vectors
    Vector2 upVector = new Vector2(0, -1);
    Vector2 motion = new Vector2(0, 0);
    // variables
    int playerDir = -1;
    int damage;
    int jumpAttackCounter = 0;
    bool isPlayerInArea = false;
    bool endOfDeathAnim = false;
    int randomAttack;
    int lastAttack;
    bool wasRoarAttackYet = false;

    int health = MAX_HEALTH;
    int direction = 1;


    //===========================

    enum States
    {
        initial,
        battlecry,
        idle,
        attackRoar,
        prepareJump,
        attackJump,
        attackRun,
        prerun,
        dash,
        hitAfterDash,
        hitNear,
        deathFall,
        death
    }
    States state;

    public override void _Ready()
    {
        longRayLeft = GetNode<RayCast2D>("RayIsPlayerLeft");
        longRayRight = GetNode<RayCast2D>("RayIsPlayerRight");
        shortRayLeft = GetNode<RayCast2D>("RayPlayerNearLeft");
        shortRayRight = GetNode<RayCast2D>("RayPlayerNearRight");
        areaStopRunAttack = GetNode<Area2D>("AreaPlayerRunAttack");

        animationPlayer = GetNode<AnimationPlayer>("AnimationPlayer");
        hitAnimationPlayer = GetNode<AnimationPlayer>("HitAnimationPlayer");
        introTimer = GetNode<Timer>("IntroTimer");
        battlecryTimer = GetNode<Timer>("BattlecryTimer");
        idleTimer = GetNode<Timer>("IdleTimer");
        dashTimer = GetNode<Timer>("DashTimer");
        jumpPrepareTimer = GetNode<Timer>("JumpPrepareTimer");
        openGateTimer = GetNode<Timer>("OpenGateTimer");
        rockSpawnTimer = GetNode<Timer>("RockSpawnTimer");

        stats = GetParent().GetNode<KinematicBody2D>("../Player").GetNode<Node>("Stats");
        playerVariables = GetNode<Node>("/root/PlayerVariables");
        camera2D = GetParent().GetNode<Camera2D>("../Player/PlayerCamera");
        playerDetection = GetParent().GetNode<Area2D>("./PlayerDetectionBossTrigger");
        swordHitbox = GetParent().GetNode<Area2D>("../Player/HitboxPivot/SwordHitbox");
        hurtboxCollision = GetNode<CollisionShape2D>("Hurtbox/CollisionShape2D");
        hitboxCollision = GetNode<CollisionShape2D>("Hitbox/CollisionShape2D");
        dustParticles = (PackedScene)ResourceLoader.Load("res://Particles/FootStepsParticle.tscn");
        fallingRock = (PackedScene)ResourceLoader.Load("res://Hazards/FallingRock/FallingRock.tscn");

        healthBar = GetParent().GetNode<ProgressBar>("./BossUI/HealthBar");
        bossNameLabel = GetParent().GetNode<Label>("./BossUI/Name");

        audioStreamPlayer = (PackedScene)ResourceLoader.Load("res://Sound/AudioStreamPlayer.tscn");
        asHit = (AudioStreamSample)ResourceLoader.Load("res://Sound/e_hit.wav");
        asDash = (AudioStreamSample)ResourceLoader.Load("res://Sound/p_dash.wav");
        asJump = (AudioStreamSample)ResourceLoader.Load("res://Sound/boss_doors_open.wav");
        asImpact = (AudioStreamSample)ResourceLoader.Load("res://Sound/boss_doors_close.wav");
        asDeath = (AudioStreamSample)ResourceLoader.Load("res://Sound/boss_death2.wav");

        playerDetection.Connect("playerEnteredBossAlertSignal", this, "onPlayerDetected");

        state = States.initial;

        bossNameLabel.Visible = false;
        healthBar.Visible = false;
        healthBar.MaxValue = MAX_HEALTH;
        healthBar.Value = healthBar.MaxValue;
    }

    public override void _PhysicsProcess(float delta)
    {
        if (state == States.initial) // INITIAL
        {
            animationPlayer.Play("IdleRight");
        }
        else if (state == States.battlecry) // BATTLECRY
        {
            animationPlayer.Play("BattlecryLeft");
        }
        else if (state == States.idle) // IDLE
        {
            // ANIMATION
            if (direction == -1) animationPlayer.Play("IdleLeft");
            else if (direction == 1) animationPlayer.Play("IdleRight");
        }
        else if (state == States.attackRun) // RUN ATTACK
        {
            // BEHAVIOUR
            motion.x += direction * ACCELERATION * delta;
            motion.x = Mathf.Clamp(motion.x, -MAX_SPEED, MAX_SPEED);

            Gravity(delta);
            motion = MoveAndSlide(motion, upVector);

            // TRANSITION TRIGGERS
            if (isPlayerInArea)
            {
                state = States.dash;
                dashTimer.Start(0.25f);
                PlaySound(asDash);

                if (direction == -1) animationPlayer.Play("DashLeft");
                else if (direction == 1) animationPlayer.Play("DashRight");
            }
            else
            {
                if (direction == -1) animationPlayer.Play("RunLeft");
                else if (direction == 1) animationPlayer.Play("RunRight");
            }

        }
        else if (state == States.dash) // DASH
        {
            // BEHAVIOUR  
            motion.x = direction * MAX_DASH_SPEED;
            motion.y = 0;
            motion = MoveAndSlide(motion, upVector);

        }
        else if (state == States.hitAfterDash) // AFTER DASH ATTACK
        {
            // BEHAVIOUR

            motion.x += direction * ACCELERATION * delta;
            motion.x = Mathf.Clamp(motion.x, -MAX_SPEED, MAX_SPEED);

            Friction(FRICTION);
            Gravity(delta);
            motion = MoveAndSlide(motion, upVector);
        }
        else if (state == States.attackJump) // JUMP ATTACK
        {
            // vertical motion
            Gravity(delta);
            motion = MoveAndSlide(motion, upVector);

            // TRANSITION TRIGGERS
            if (IsOnFloor())
            {
                ToStatePrepareJumpTransition();
                MakeDustParticle();
                PlaySound(asImpact);
            }
        }
        else if (state == States.hitNear) // HIT NEAR
        {
            // vertical motion
            motion.x = direction * 50;
            motion = MoveAndSlide(motion, upVector);

        }
        else if (state == States.deathFall) // DEATH FALL
        {
            Gravity(delta);
            Friction(0.01f);
            motion = MoveAndSlide(motion, upVector);

            if (IsOnFloor() && endOfDeathAnim)
            {
                if (direction == -1) animationPlayer.Play("CorpseLeft");
                else if (direction == 1) animationPlayer.Play("CorpseRight");
                state = States.death;
                FromDeathFallToDeathState();
            }
        }
    }

    public void Gravity(float delta)
    {
        motion.y = Mathf.Min(motion.y + GRAVITY * delta, MAX_FALL_SPEED);
    }

    public void Friction(float friction)
    {
        motion.x = Mathf.Lerp(motion.x, 0, friction);
    }

    public void PlaySound(AudioStreamSample sample)
    {
        AudioStreamPlayer asp = (AudioStreamPlayer)audioStreamPlayer.Instance();
        asp.Stream = sample;
        GetTree().CurrentScene.AddChild(asp);
    }

    public void TripleRockSpawn()
    {
        int x1 = r.Next(29, 171);
        int x2 = r.Next(172, 310);
        int x3 = r.Next(311, 452);

        RockSpawn(x1);
        RockSpawn(x2);
        RockSpawn(x3);
    }

    public void RockSpawn(int x)
    {
        KinematicBody2D rock = (KinematicBody2D)fallingRock.Instance();
        rock.GlobalPosition = new Vector2(x, 48);
        GetTree().CurrentScene.CallDeferred("add_child", rock);
    }

    public void MakeDustParticle(int x1 = 0, int y1 = 0)
    {
        Particles2D dust = (Particles2D)dustParticles.Instance();
        dust.GlobalPosition = new Vector2(GlobalPosition.x + x1, GlobalPosition.y + y1);
        dust.Emitting = true;
        GetParent().AddChild(dust);
    }

    public int GetPlayerDirection()
    {
        if (longRayRight.IsColliding())
        {
            return 1;
        }
        else if (longRayLeft.IsColliding())
        {
            return -1;
        }
        else
        {
            return 0;
        }
    }

    public int GetNearbyPlayerDirection()
    {
        if (shortRayRight.IsColliding())
        {
            return 1;
        }
        else if (shortRayLeft.IsColliding())
        {
            return -1;
        }
        else
        {
            return 0;
        }
    }


    public void EndOfAttack()
    {
        state = States.idle;
        idleTimer.Start(2);
    }

    public void onJumpPrepareTimerTimeout()
    {
        if (jumpAttackCounter >= 3)
        {
            jumpAttackCounter = 0;
            state = States.idle;
            idleTimer.Start(0.5f);
        }
        else
        {
            if ((playerDir = GetPlayerDirection()) != 0)
            {
                jumpAttackCounter++;
                direction = playerDir;
                motion.y = -JUMP_FORCE;
                motion.x = (playerBody.GlobalPosition.x - this.GlobalPosition.x) / AIR_TIME;
                state = States.attackJump;
                MakeDustParticle();
                PlaySound(asJump);
                // ANIMATIONS
                if (direction == -1) animationPlayer.Play("JumpLeft");
                else if (direction == 1) animationPlayer.Play("JumpRight");
            }
            else
            {
                jumpPrepareTimer.Start(0.1f);
            }
        }

    }


    public void onIntroTimerTimeout()
    {
        direction = -1;
        state = States.battlecry;
        battlecryTimer.Start(BATTLECRY_TIME);
        camera2D.Call("StartShake", BATTLECRY_TIME, 0.02f, SHAKE_STRENGTH, 2);
        playerBody = (KinematicBody2D)longRayLeft.GetCollider();

        bossNameLabel.Visible = true;
        healthBar.Visible = true;
    }

    public void onBattlecryTimerTimeout()
    {
        state = States.idle;
        idleTimer.Start(2);
        rockSpawnTimer.Stop();
    }

    // CHOOSING THE ATTACK
    public void onIdleTimerTimeout()
    {
        if ((playerDir = GetNearbyPlayerDirection()) != 0)
        {
            direction = playerDir;
            state = States.hitNear;
            if (direction == -1) animationPlayer.Play("HitNearLeft");
            else if (direction == 1) animationPlayer.Play("HitNearRight");
        }
        else if ((playerDir = GetPlayerDirection()) != 0)
        {
            direction = playerDir;

            if (health <= (MAX_HEALTH / 2) && !wasRoarAttackYet)
            {
                wasRoarAttackYet = true;
                ToStateRoarAttackTransition();
            }
            else
            {
                randomAttack = r.Next(0, 4);

                if (randomAttack <= 1)
                {
                    Prerun();
                    lastAttack = 0;
                }
                else if (randomAttack == 2)
                {
                    ToStatePrepareJumpTransition();
                    lastAttack = 1;
                }
                else if (randomAttack == 3)
                {
                    if (health > (MAX_HEALTH / 2))
                    {
                        Prerun();
                        lastAttack = 0;
                    }
                    else
                    {
                        if (lastAttack != 2)
                        {
                            ToStateRoarAttackTransition();
                        }
                        else
                        {
                            randomAttack = r.Next(0, 2);
                            if (randomAttack == 0)
                            {
                                Prerun();
                                lastAttack = 0;
                            }
                            else if (randomAttack == 1)
                            {
                                ToStatePrepareJumpTransition();
                                lastAttack = 1;
                            }
                        }
                    }
                }
            }

        }
        else
        {
            idleTimer.Start(0.1f);
        }
    }

    public void ToStatePrepareJumpTransition()
    {
        state = States.prepareJump;
        jumpPrepareTimer.Start(0.3f);
        if (direction == -1) animationPlayer.Play("JumpPrepareLeft");
        else if (direction == 1) animationPlayer.Play("JumpPrepareRight");
    }


    public void ToStateRoarAttackTransition()
    {
        state = States.attackRoar;
        battlecryTimer.Start(ROAR_ATTACK_TIME);
        camera2D.Call("StartShake", ROAR_ATTACK_TIME, 0.02f, SHAKE_STRENGTH, 2);
        rockSpawnTimer.Start(0.2f);
        lastAttack = 2;
        if (direction == -1) animationPlayer.Play("BattlecryLeft");
        else if (direction == 1) animationPlayer.Play("BattlecryRight");
    }


    public void onDashTimerTimeout()
    {
        state = States.hitAfterDash;

        if (direction == -1) animationPlayer.Play("HitAfterDashLeft");
        else if (direction == 1) animationPlayer.Play("HitAfterDashRight");
    }

    public void onRockSpawnTimerTimeout()
    {
        TripleRockSpawn();
    }

    public void onOpenGateTimerTimeout()
    {
        this.EmitSignal(nameof(bossKilledOpenTheGateSignal));
    }


    public void onAreaPlayerBodyEntered(KinematicBody2D player)
    {
        isPlayerInArea = true;
    }

    public void onAreaPlayerBodyExited(KinematicBody2D player)
    {
        isPlayerInArea = false;
    }

    public void onPlayerDetected() // intro trigger area
    {
        introTimer.Start(2);
    }


    public void onHurtboxAreaEntered(Area2D area)
    {
        damage = (int)swordHitbox.Get("damage");

        health -= damage;
        this.hitAnimationPlayer.Play("Start");
        PlaySound(asHit);

        healthBar.Call("SetBarValue", health);

        if (health <= 0)
        {
            playerDir = (int)playerBody.Get("direction");
            direction = -playerDir;
            if (direction == -1) animationPlayer.Play("DeathLeft");
            else if (direction == 1) animationPlayer.Play("DeathRight");

            hurtboxCollision.SetDeferred("disabled", true);
            hitboxCollision.SetDeferred("disabled", true);

            motion.y = 0;
            motion.x = -direction * 300;
            stats.Set("Souls", (int)stats.Get("Souls") + SOUL);
            stats.Set("Kills", (int)stats.Get("Kills") + 1);
            state = States.deathFall;
            Engine.TimeScale = 0.2f;
            PlaySound(asHit);
            idleTimer.Stop();
            dashTimer.Stop();
            rockSpawnTimer.Stop();
            jumpPrepareTimer.Stop();
            battlecryTimer.Stop();

            bossNameLabel.Visible = false;
            healthBar.Visible = false;
            playerVariables.Set("Level", (int)playerVariables.Get("Level") + 1);
            playerVariables.Call("SaveData");
        }
    }

    public void FromDeathFallToDeathState()
    {
        MakeDustParticle(-24, 0);
        MakeDustParticle(17, 0);

        Engine.TimeScale = 1;
        PlaySound(asDeath);
        openGateTimer.Start(2);
    }

    public void FromPrerunToRunAttack()
    {
        state = States.attackRun;
    }

    public void Prerun()
    {
        state = States.prerun;
        if (direction == -1) animationPlayer.Play("PrerunLeft");
        else if (direction == 1) animationPlayer.Play("PrerunRight");
    }

    public void EndOfDeathAnim()
    {
        endOfDeathAnim = true;
    }

}