using Godot;
using System;

public class Fleye : KinematicBody2D
{
    // nodes
    Node stats;
    PlayerVariables playerVariables;
    AnimationPlayer animationPlayer;
    AnimationPlayer hitAnimationPlayer;
    KinematicBody2D player;
    Area2D swordHitbox;
    Area2D playerDetectionZone;
    CollisionShape2D hurtboxCollision;
    CollisionShape2D hitboxCollision;
    Timer lookingAroundTimer;
    PackedScene audioStreamPlayer;
    AudioStreamSample asHit;
    AudioStreamSample asAlert;
    // constants
    const int ACCELERATION = 300;
    const int MAX_SPEED = 80;
    const int KNOCKBACK = 200;
    const int MAX_FALL_SPEED = 500;
    const int GRAVITY = 600;
    const float FRICTION = 0.1f;
    const int SOUL = 20;
    // vectors
    Vector2 upVector = new Vector2(0, -1);
    Vector2 motion = new Vector2(0, 0);
    Vector2 startingPosition = new Vector2(0, 0);
    // variables
    int playerDirection;
    int damage;
    bool dead = false;

    int health = 2;
    int direction = -1;

    enum States
    {
        idle,
        chase,
        returning,
        deathAir,
        deathFall,
        deathGround
    }
    States state;

    public override void _Ready()
    {
        animationPlayer = GetNode<AnimationPlayer>("AnimationPlayer");
        hitAnimationPlayer = GetNode<AnimationPlayer>("HitAnimationPlayer");

        player = GetParent().GetNode<KinematicBody2D>("../Player");
        stats = player.GetNode<Node>("Stats");
        swordHitbox = GetParent().GetNode<Area2D>("../Player/HitboxPivot/SwordHitbox");
        playerDetectionZone = GetNode<Area2D>("PlayerDetectionZone");
        hurtboxCollision = GetNode<CollisionShape2D>("Hurtbox/CollisionShape2D");
        hitboxCollision = GetNode<CollisionShape2D>("Hitbox/CollisionShape2D");
        lookingAroundTimer = GetNode<Timer>("LookingAroundTimer");
        playerVariables = GetNode<PlayerVariables>("/root/PlayerVariables");

        audioStreamPlayer = (PackedScene)ResourceLoader.Load("res://Sound/AudioStreamPlayer.tscn");
        asHit = (AudioStreamSample)ResourceLoader.Load("res://Sound/e_hit.wav");
        asAlert = (AudioStreamSample)ResourceLoader.Load("res://Sound/fleye_alert.wav");

        startingPosition = GlobalPosition;
        direction = -1;

        state = States.idle;
    }


    public override void _PhysicsProcess(float delta)
    {
        if (state == States.idle) // IDLE
        {
            // BEHAVIOUR

            // TRANSITIONS TRIGGERS
            SeekPlayer(); // if player detected - change state to chase

            // ANIMATION
            if (direction <= -1) animationPlayer.Play("FlyingLeft");
            else animationPlayer.Play("FlyingRight");
        }
        else if (state == States.chase) // CHASE
        {
            // BEHAVIOUR
            KinematicBody2D playerBody = (KinematicBody2D)playerDetectionZone.Get("player");
            if (playerBody != null)
            {
                var course = (playerBody.GlobalPosition - GlobalPosition).Normalized();
                motion = motion.MoveToward(course * MAX_SPEED, ACCELERATION * delta);
                direction = Mathf.Sign(course.x);
            }
            motion = MoveAndSlide(motion, upVector);

            // TRANSITIONS TRIGGERS
            if (playerBody == null)
            {
                state = States.returning;
            }

            // ANIMATION
            if (direction <= -1) animationPlayer.Play("FlyingLeft");
            else animationPlayer.Play("FlyingRight");

        }
        else if (state == States.returning) // RETURNING
        {
            // BEHAVIOUR
            var course = (startingPosition - GlobalPosition).Normalized();
            motion = motion.MoveToward(course * MAX_SPEED, ACCELERATION * delta);
            motion = MoveAndSlide(motion, upVector);

            direction = Mathf.Sign(course.x);

            // TRANSITIONS TRIGGERS
            SeekPlayer(); // if player detected - change state to chase

            if (Mathf.Abs(startingPosition.y - GlobalPosition.y) < 5)
            {
                state = States.idle;
            }

            // ANIMATION
            if (direction <= -1) animationPlayer.Play("FlyingLeft");
            else animationPlayer.Play("FlyingRight");

        }
        else if (state == States.deathAir) // DEATH AIR
        {
            // BEHAVIOUR
            Gravity(delta);
            motion = MoveAndSlide(motion, upVector);

            // TRANSITIONS TRIGGERS
            if (!animationPlayer.IsPlaying())
            {
                state = States.deathFall;
            }

            if (!dead)
            {
                dead = true;
                stats.Set("Souls", (int)stats.Get("Souls") + SOUL);
                stats.Set("Kills", (int)stats.Get("Kills") + 1);
            }


        }
        else if (state == States.deathFall) // DEATH FALL
        {
            // BEHAVIOUR
            Friction(0.3f);
            Gravity(delta);
            motion = MoveAndSlide(motion, upVector);

            // TRANSITIONS TRIGGERS
            if (IsOnFloor())
            {
                state = States.deathGround;
                if (direction <= -1) animationPlayer.Play("DeathImpactLeft");
                else animationPlayer.Play("DeathImpactRight");
            }
            else
            {
                if (direction <= -1) animationPlayer.Play("DeathFallLeft");
                else animationPlayer.Play("DeathFallRight");
            }

            // ANIMATION

        }
        else if (state == States.deathGround) // CORPSE
        {
            // BEHAVIOUR
            Friction(0.3f);
            Gravity(delta);
            motion = MoveAndSlide(motion, upVector);

        }
    }

    public void SeekPlayer()
    {
        bool canSeePlayer = (bool)playerDetectionZone.Call("CanSeePlayer");
        if (canSeePlayer)
        {
            PlaySound(asAlert);
            state = States.chase;
        }
    }

    public void Gravity(float delta)
    {
        motion.y = Mathf.Min(motion.y + GRAVITY * delta, MAX_FALL_SPEED);
    }

    public void Friction(float friction)
    {
        motion.x = Mathf.Lerp(motion.x, 0, friction);
    }

    public void PlaySound(AudioStreamSample sample)
    {
        AudioStreamPlayer asp = (AudioStreamPlayer)audioStreamPlayer.Instance();
        asp.Stream = sample;
        GetTree().CurrentScene.AddChild(asp);
    }

    public void onHurtboxAreaEntered(Area2D area)
    {
        playerDirection = (int)player.Get("direction");
        damage = (int)swordHitbox.Get("damage");

        health -= damage;
        motion.x = playerDirection * KNOCKBACK;
        this.hitAnimationPlayer.Play("Start");
        PlaySound(asHit);

        if (health > 0)
        {
            direction = -playerDirection;
        }
        else if (health <= 0)
        {
            direction = -playerDirection;
            if (direction <= -1) animationPlayer.Play("DeathLeft");
            else animationPlayer.Play("DeathRight");

            hurtboxCollision.SetDeferred("disabled", true);
            hitboxCollision.SetDeferred("disabled", true);

            state = States.deathAir;
        }
    }

    public void onLookingAroundTimerTimeout()
    {
        if (state == States.idle)
        {
            direction = -direction;
        }
        else if (state == States.deathGround)
        {
            lookingAroundTimer.Stop();
        }
    }

}
