using Godot;
using System;
using System.Text.RegularExpressions;

public class Player : KinematicBody2D
{
    // nodes
    Node playerVariables;
    Node stats;
    Sprite playerSprite;
    AnimationPlayer animationPlayer;
    AnimationPlayer hitAnimationPlayer;
    RayCast2D leftWallRayCast1;
    RayCast2D leftWallRayCast2;
    RayCast2D rightWallRayCast1;
    RayCast2D rightWallRayCast2;
    Area2D hurtbox;
    CollisionShape2D hurtboxCollision;
    CollisionShape2D hitboxCollision;
    WorldEnvironment worldEnvironment;
    AnimationPlayer worldAnimationPlayer;
    Camera2D camera2D;
    Position2D respawnPosition;
    CollisionShape2D respawnCollisionShape;
    Timer spikeDeathTimer;
    Timer deathTransitionTimer;
    Timer transitionTimer;
    Timer ghostTimer;
    PackedScene dashGhost;
    PackedScene dustParticles;
    // audio
    PackedScene audioStreamPlayer;
    AudioStreamSample asJump;
    AudioStreamSample asImpact;
    AudioStreamSample asDash;
    AudioStreamSample asHurt;
    AudioStreamSample asDeath;
    AudioStreamSample asPickup;
    AudioStreamSample asImpactPrepared;
    AudioStreamSample asLost;
    // colors
    Color blueParticle = new Color("db606480");
    Color redParticle = new Color("db806060");
    Color goldParticle = new Color("db8a835a");
    // constants
    const int ACCELERATION = 1000;
    const int WALL_SLIDE_ACCELERATION = 5;
    const int MAX_SPEED = 100;
    const int MAX_FALL_SPEED = 300;
    const int MAX_DASH_SPEED = 350;
    const float FRICTION = 0.3f;
    const float AIR_FRICTION = 0.2f;
    const int GRAVITY = 600;
    const int JUMP_FORCE = 260;
    const int WALL_JUMP_FORCE = 220;
    const int MAX_JUMPS = 1;
    const int MAX_WALL_SLIDE_SPEED = 50;
    const int CAYOTE_TIME = 8;
    const int CAYOTE_WJ_TIME = 16;
    const int JUMP_BUFF_TIME = 8;
    const int DASH_TIME = 12 + DASH_KEY_WAIT_TIME;
    const int DASH_RELOAD_TIME = 10;
    const int DASH_BUFF_TIME = 8;
    const int DASH_KEY_WAIT_TIME = 3;
    const int WALL_JUMP_TIME = 6;
    const int WALL_JUMP_BUFF_TIME = 20;
    const int INVINCIBILITY_TIME = 2;
    const int LAST_LEVEL = 25;
    // vectors
    Vector2 motion = new Vector2(0, 0);
    Vector2 upVector = new Vector2(0, -1);
    // booleans
    bool longJump = false;
    bool dashReloaded = true;
    bool attackPressed = false;
    bool onLeftWall = false;
    bool leftWallDrop;
    bool rightWallDrop;
    bool wasOnFloor;
    bool spikesTouched = false;
    bool paused = false;
    // timers
    int cayoteTimer = 0;
    int jumpBuffTimer = 0;
    int dashTimer = 0;
    int dashReloadTimer = 0;
    int dashBuffTimer = 0;
    int dashKeyWaitTimer = 0;
    int wallJumpTimer = 0;
    int leftWJBuffTimer = 0;
    int rightWJBuffTimer = 0;
    // Player abilities
    bool canDash;
    bool canWallJump;
    // variables
    float x_input = 1f;
    int direction = 1;
    int attackIndex = 0;

    // Statistics
    int currentLevel;

    [Signal] public delegate void lastLevelResetSignal();
    [Signal] public delegate void fromPreparedState();

    // ========================================

    enum States
    {
        idle,
        run,
        air,
        crouch,
        attack,
        hurt,
        death,
        dash,
        wallSlide,
        prepared,
        noInput,
        freeze
    }
    States state;

    public override void _Ready()
    {
        playerVariables = GetNode<Node>("/root/PlayerVariables");
        stats = GetNode<Node>("Stats");
        playerSprite = GetNode<Sprite>("Sprite");
        animationPlayer = GetNode<AnimationPlayer>("AnimationPlayer");
        hitAnimationPlayer = GetNode<AnimationPlayer>("HitAnimationPlayer");
        leftWallRayCast1 = GetNode<RayCast2D>("LeftWallRayCast1");
        leftWallRayCast2 = GetNode<RayCast2D>("LeftWallRayCast2");
        rightWallRayCast1 = GetNode<RayCast2D>("RightWallRayCast1");
        rightWallRayCast2 = GetNode<RayCast2D>("RightWallRayCast2");
        hitboxCollision = GetNode<CollisionShape2D>("HitboxPivot/SwordHitbox/CollisionShape2D");
        hurtboxCollision = GetNode<CollisionShape2D>("Hurtbox/CollisionShape2D");
        hurtbox = GetNode<Area2D>("Hurtbox");
        worldEnvironment = GetNode<WorldEnvironment>("../WorldEnvironment");
        worldAnimationPlayer = GetNode<AnimationPlayer>("../WorldAnimationPlayer");
        camera2D = GetNode<Camera2D>("./PlayerCamera");
        respawnPosition = GetNode<Position2D>("../RespawnPosition");
        respawnCollisionShape = GetNode<CollisionShape2D>("../RespawnPosition/RespawnArea/RespawnCollisionShape");
        spikeDeathTimer = GetNode<Timer>("Timers/SpikeDeathTimer");
        deathTransitionTimer = GetNode<Timer>("Timers/DeathTransitionTimer");
        transitionTimer = GetNode<Timer>("Timers/TransitionTimer");
        ghostTimer = GetNode<Timer>("Timers/GhostTimer");
        dashGhost = (PackedScene)ResourceLoader.Load("res://Player/DashGhost.tscn");
        dustParticles = (PackedScene)ResourceLoader.Load("res://Particles/FootStepsParticle.tscn");

        // audio
        audioStreamPlayer = (PackedScene)ResourceLoader.Load("res://Sound/AudioStreamPlayer.tscn");
        asJump = (AudioStreamSample)ResourceLoader.Load("res://Sound/p_jump3.wav");
        asImpact = (AudioStreamSample)ResourceLoader.Load("res://Sound/p_footstep2.wav");
        asDash = (AudioStreamSample)ResourceLoader.Load("res://Sound/p_dash.wav");
        asHurt = (AudioStreamSample)ResourceLoader.Load("res://Sound/p_hurt2.wav");
        asDeath = (AudioStreamSample)ResourceLoader.Load("res://Sound/p_death.wav");
        asImpactPrepared = (AudioStreamSample)ResourceLoader.Load("res://Sound/boss_doors_close.wav");
        asLost = (AudioStreamSample)ResourceLoader.Load("res://Sound/challenge_lost.wav");

        stats.Connect("noHealthSignal", this, "onNoHealthSignal");

        worldEnvironment.Environment.AdjustmentSaturation = 1;
        worldEnvironment.Environment.DofBlurNearTransition = 5;
        worldEnvironment.Environment.DofBlurNearDistance = 1;

        if ((bool)playerVariables.Get("makeTransition"))
        {
            playerVariables.Set("makeTransition", false);
            camera2D.Call("FadeOut");
        }

        // Get global variables
        canDash = (bool)playerVariables.Get("canDash");
        canWallJump = (bool)playerVariables.Get("canWallJump");

        currentLevel = GetLevelNumber();
        playerVariables.Set("Level", currentLevel);
        playerVariables.Call("SaveData");

        hitAnimationPlayer.Play("Stop");

        if (currentLevel == 1)
        {
            wasOnFloor = false;
            state = States.prepared;
        }
        else
        {
            state = States.idle;
        }

        if ((bool)playerVariables.Get("loaded"))
        {
            playerVariables.Set("loaded", false);
            if (currentLevel != 1)
            {
                GlobalPosition = respawnPosition.GlobalPosition;
                state = States.prepared;
            }
        }
    }

    //  // Called every frame. 'delta' is the elapsed time since the previous frame.
    public override void _PhysicsProcess(float delta)
    {
        if (Input.IsActionJustPressed("ui_cancel"))
        {
            GetTree().ChangeScene("res://Menu/MainMenu.tscn");
        }


        // TIMERS
        jumpBuffTimer = UpdateTimer(jumpBuffTimer);
        cayoteTimer = UpdateTimer(cayoteTimer);
        dashTimer = UpdateTimer(dashTimer);
        dashReloadTimer = UpdateTimer(dashReloadTimer);
        dashBuffTimer = UpdateTimer(dashBuffTimer);
        leftWJBuffTimer = UpdateTimer(leftWJBuffTimer);
        rightWJBuffTimer = UpdateTimer(rightWJBuffTimer);
        dashKeyWaitTimer = UpdateTimer(dashKeyWaitTimer);
        wallJumpTimer = UpdateTimer(wallJumpTimer);

        // INPUT CHECK
        x_input = Input.GetActionStrength("key_right") - Input.GetActionStrength("key_left");
        if (state != States.dash && state != States.attack && state != States.wallSlide && wallJumpTimer == 0 && x_input != 0) direction = (int)x_input;

        // STATE MACHINE
        if (state == States.idle) // IDLE
        {
            // BEHAVIOUR
            Friction(FRICTION);
            Gravity(delta);
            motion = MoveAndSlide(motion, upVector);

            // TRANSITION TRIGGERS
            if (Input.IsActionPressed("key_down"))
            {
                state = States.crouch;
            }
            else if (Input.IsActionJustPressed("key_attack"))
            {
                state = States.attack;
            }
            else if (dashReloadTimer == 0 && (Input.IsActionJustPressed("key_dash") || dashBuffTimer > 0) && canDash)
            {
                MakeDustParticle();
                dashTimer = DASH_TIME;
                dashBuffTimer = 0;
                dashKeyWaitTimer = DASH_KEY_WAIT_TIME;
                PlaySound(asDash);
                camera2D.Call("StartShake", 0.3f, 0.01f, 1f, 1);
                state = States.dash;
            }
            else if (Input.IsActionJustReleased("key_jump"))
            {
                jumpBuffTimer = 0;
            }
            else if (Input.IsActionJustPressed("key_jump") || jumpBuffTimer > 0)
            {
                MakeDustParticle();
                motion.y = -JUMP_FORCE;
                PlaySound(asJump);
                jumpBuffTimer = 0;
                if (Input.IsActionPressed("key_jump")) longJump = true;
                state = States.air;
            }
            else if (x_input != 0 && !(x_input == 1 && NextToRightWall()) && !(x_input == -1 && NextToLeftWall()))
            {
                state = States.run;
            }
            else if (!IsOnFloor())
            {
                state = States.air;
            }

            // ANIMATION
            if (direction == 1) animationPlayer.Play("IdleAlertRight");
            else if (direction == -1) animationPlayer.Play("IdleAlertLeft");
        }
        else if (state == States.run) // RUN
        {
            // BEHAVIOUR
            motion.x += direction * ACCELERATION * delta;
            motion.x = Mathf.Clamp(motion.x, -MAX_SPEED, MAX_SPEED);

            Gravity(delta);
            motion = MoveAndSlide(motion, upVector);

            // TRANSITION TRIGGERS
            if (Input.IsActionPressed("key_down"))
            {
                state = States.crouch;
            }
            else if (Input.IsActionJustPressed("key_attack"))
            {
                state = States.attack;
            }
            else if (dashReloadTimer == 0 && (Input.IsActionJustPressed("key_dash") || dashBuffTimer > 0) && canDash)
            {
                MakeDustParticle();
                dashTimer = DASH_TIME;
                dashBuffTimer = 0;
                dashKeyWaitTimer = DASH_KEY_WAIT_TIME;
                PlaySound(asDash);
                camera2D.Call("StartShake", 0.3f, 0.01f, 1f, 1);
                state = States.dash;
            }
            else if (Input.IsActionJustPressed("key_jump") || jumpBuffTimer > 0)
            {
                MakeDustParticle();
                motion.y = -JUMP_FORCE;
                PlaySound(asJump);
                jumpBuffTimer = 0;
                if (Input.IsActionPressed("key_jump")) longJump = true;
                state = States.air;
            }
            else if (x_input == 0 || (x_input == 1 && NextToRightWall()) || (x_input == -1 && NextToLeftWall()))
            {
                state = States.idle;
            }
            else if (!IsOnFloor())
            {
                cayoteTimer = CAYOTE_TIME;
                state = States.air;
            }

            // ANIMATION
            if (direction == 1) animationPlayer.Play("RunRight");
            else if (direction == -1) animationPlayer.Play("RunLeft");
        }
        else if (state == States.air) // AIR
        {
            // BEHAVIOUR

            // horizontal motion
            if (wallJumpTimer > 0) // pokud je právě vykonáván walljump, hráč se pohybuje směrem od zdi
            {
                motion.x = direction * MAX_SPEED * 1.5f;
            }
            else
            {
                if (x_input == 0)
                {
                    Friction(AIR_FRICTION);
                }
                else
                {
                    motion.x += direction * (ACCELERATION * 1.5f) * delta;
                    motion.x = Mathf.Clamp(motion.x, -MAX_SPEED, MAX_SPEED);
                }
            }

            // vertical motion
            if (Input.IsActionJustPressed("key_jump"))
            {
                if (cayoteTimer > 0)
                {
                    motion.y = -JUMP_FORCE;
                    PlaySound(asJump);
                    longJump = true;
                    cayoteTimer = 0;
                }
                else
                {
                    jumpBuffTimer = JUMP_BUFF_TIME;
                }
            }
            else if (Input.IsActionJustReleased("key_jump"))
            {
                longJump = false;
                wallJumpTimer = 0;
                if (motion.y < 0)
                {
                    motion.y = 0.3f * motion.y;
                }
                else
                {
                    jumpBuffTimer = 0;
                }
            }

            if (Input.IsActionJustPressed("key_dash")) dashBuffTimer = DASH_BUFF_TIME;
            if (Input.IsActionJustReleased("key_left")) leftWJBuffTimer = WALL_JUMP_BUFF_TIME;
            if (Input.IsActionJustReleased("key_right")) rightWJBuffTimer = WALL_JUMP_BUFF_TIME;

            if (wallJumpTimer == 0) Gravity(delta);
            motion = MoveAndSlide(motion, upVector);

            // TRANSITION TRIGGERS
            if (Input.IsActionJustPressed("key_attack") && wallJumpTimer == 0)
            {
                state = States.attack;
            }
            else if (Input.IsActionJustPressed("key_dash") && dashReloadTimer == 0 && dashReloaded && canDash)
            {
                if (Input.IsActionJustPressed("key_left")) direction = -1;
                else if (Input.IsActionJustPressed("key_right")) direction = 1;
                longJump = false;
                dashTimer = DASH_TIME;
                dashKeyWaitTimer = DASH_KEY_WAIT_TIME;
                PlaySound(asDash);
                camera2D.Call("StartShake", 0.3f, 0.01f, 1f, 1);
                state = States.dash;
            }
            else if (canWallJump && ((Input.IsActionJustPressed("key_jump") || (jumpBuffTimer > 0)) && NextToLeftWall() && leftWJBuffTimer > 0))
            {
                motion.x = JUMP_FORCE;
                motion.y = -WALL_JUMP_FORCE;
                PlaySound(asJump);
                wallJumpTimer = WALL_JUMP_TIME;
                jumpBuffTimer = 0;
                leftWJBuffTimer = 0;
                direction = -direction;
                if (Input.IsActionPressed("key_jump")) longJump = true;
                state = States.air;
            }
            else if (canWallJump && ((Input.IsActionJustPressed("key_jump") || (jumpBuffTimer > 0)) && NextToRightWall() && rightWJBuffTimer > 0))
            {
                motion.x = -JUMP_FORCE;
                motion.y = -WALL_JUMP_FORCE;
                PlaySound(asJump);
                wallJumpTimer = WALL_JUMP_TIME;
                jumpBuffTimer = 0;
                rightWJBuffTimer = 0;
                direction = -direction;
                if (Input.IsActionPressed("key_jump")) longJump = true;
                state = States.air;
            }
            else if (NextToLeftWall() && Input.IsActionPressed("key_left") && wallJumpTimer == 0 && canWallJump)
            {
                longJump = false;
                dashReloaded = true;
                onLeftWall = true;
                state = States.wallSlide;
            }
            else if (NextToRightWall() && Input.IsActionPressed("key_right") && wallJumpTimer == 0 && canWallJump)
            {
                longJump = false;
                dashReloaded = true;
                onLeftWall = false;
                state = States.wallSlide;
            }
            else if (IsOnFloor())
            {
                longJump = false;
                dashReloaded = true;
                MakeDustParticle();
                PlaySound(asImpact);

                if (Input.IsActionPressed("key_down"))
                {
                    state = States.crouch;
                }
                else if (x_input == 0)
                {
                    state = States.idle;
                }
                else
                {
                    state = States.run;
                }
            }

            // ANIMATION
            if (!((animationPlayer.CurrentAnimation == "AirRollLeft" || animationPlayer.CurrentAnimation == "AirRollRight") && animationPlayer.IsPlaying()))
            {
                if (Mathf.Abs(motion.y) < 50 && longJump)
                {
                    if (direction == 1) animationPlayer.Play("AirRollRight");
                    else if (direction == -1) animationPlayer.Play("AirRollLeft");
                }
                else if (motion.y < 0)
                {
                    if (direction == 1) animationPlayer.Play("AirUpRight");
                    else if (direction == -1) animationPlayer.Play("AirUpLeft");
                }
                else if (Mathf.Abs(motion.y) < 90)
                {
                    if (direction == 1) animationPlayer.Play("AirMidRight");
                    else if (direction == -1) animationPlayer.Play("AirMidLeft");
                }
                else
                {
                    if (direction == 1) animationPlayer.Play("AirFallRight");
                    else if (direction == -1) animationPlayer.Play("AirFallLeft");
                }
            }


        }
        else if (state == States.dash) // DASH
        {
            // BEHAVIOUR  
            if (dashKeyWaitTimer == 0)
            {
                motion.x = direction * MAX_DASH_SPEED;
                motion.y = 0;
                motion = MoveAndSlide(motion, upVector);
            }
            else if (x_input != 0)
            {
                direction = (int)x_input;
            }

            if (Input.IsActionJustPressed("key_jump")) jumpBuffTimer = JUMP_BUFF_TIME;
            if (Input.IsActionJustReleased("key_left")) leftWJBuffTimer = WALL_JUMP_BUFF_TIME;
            if (Input.IsActionJustReleased("key_right")) rightWJBuffTimer = WALL_JUMP_BUFF_TIME;

            // TRANSITION TRIGGERS
            if (NextToLeftWall() && Input.IsActionPressed("key_left") && direction == -1 && canWallJump)
            {
                onLeftWall = true;
                state = States.wallSlide;
            }
            else if (NextToRightWall() && Input.IsActionPressed("key_right") && direction == 1 && canWallJump)
            {
                onLeftWall = false;
                state = States.wallSlide;
            }
            else if (canWallJump && ((Input.IsActionJustPressed("key_jump") || (jumpBuffTimer > 0)) && NextToLeftWall()))
            {
                motion.x = JUMP_FORCE;
                motion.y = -WALL_JUMP_FORCE;
                PlaySound(asJump);
                wallJumpTimer = WALL_JUMP_TIME;
                jumpBuffTimer = 0;
                leftWJBuffTimer = 0;
                direction = -direction;
                if (Input.IsActionPressed("key_jump")) longJump = true;
                state = States.air;
            }
            else if (canWallJump && ((Input.IsActionJustPressed("key_jump") || (jumpBuffTimer > 0)) && NextToRightWall()))
            {
                motion.x = -JUMP_FORCE;
                motion.y = -WALL_JUMP_FORCE;
                PlaySound(asJump);
                wallJumpTimer = WALL_JUMP_TIME;
                jumpBuffTimer = 0;
                rightWJBuffTimer = 0;
                direction = -direction;
                if (Input.IsActionPressed("key_jump")) longJump = true;
                state = States.air;
            }
            else if (dashTimer == 0)
            {
                dashReloaded = true;
                dashReloadTimer = DASH_RELOAD_TIME;

                if (Input.IsActionJustPressed("key_attack"))
                {
                    state = States.attack;
                }
                else if (IsOnFloor())
                {
                    if (Input.IsActionPressed("key_down"))
                    {
                        state = States.crouch;
                    }
                    else if (x_input == 0)
                    {
                        state = States.idle;
                    }
                    else
                    {
                        state = States.run;
                    }
                }
                else
                {
                    dashReloaded = false;
                    state = States.air;
                }
            }

            // ANIMATION
            if (direction == 1) animationPlayer.Play("DashRight");
            else if (direction == -1) animationPlayer.Play("DashLeft");

        }
        else if (state == States.wallSlide) // WALL SLIDE
        {
            // BEHAVIOUR
            motion.y = Mathf.Min(motion.y + WALL_SLIDE_ACCELERATION * delta, MAX_WALL_SLIDE_SPEED);

            if (Input.IsActionJustReleased("key_jump"))
            {
                if (motion.y < 0)
                {
                    motion.y = 0.3f * motion.y;
                }
            }

            Gravity(delta);
            motion = MoveAndSlide(motion, upVector);

            // TRANSITION TRIGGERS
            leftWallDrop = onLeftWall && Input.IsActionJustReleased("key_left");        // left wall, left key released
            rightWallDrop = !onLeftWall && Input.IsActionJustReleased("key_right");     // right wall, right key released

            if (Input.IsActionJustPressed("key_dash") && canDash)
            {
                direction = -direction;
                dashTimer = DASH_TIME;
                PlaySound(asDash);
                camera2D.Call("StartShake", 0.3f, 0.01f, 1f, 1);
                state = States.dash;
            }
            else if (leftWallDrop || rightWallDrop || !NextToWall())
            {
                if (leftWallDrop) direction = 1;
                else if (rightWallDrop) direction = -1;
                if (motion.y < 0) longJump = true;
                cayoteTimer = CAYOTE_WJ_TIME;
                dashReloadTimer = 0;
                dashReloaded = true;
                state = States.air;
            }
            else if (Input.IsActionJustPressed("key_jump") || jumpBuffTimer > 0)
            {
                if (onLeftWall) motion.x = JUMP_FORCE;
                else motion.x = -JUMP_FORCE;

                motion.y = -WALL_JUMP_FORCE;
                PlaySound(asJump);
                wallJumpTimer = WALL_JUMP_TIME;
                jumpBuffTimer = 0;
                direction = -direction;
                if (Input.IsActionPressed("key_jump")) longJump = true;
                state = States.air;
            }
            else if (IsOnFloor())
            {
                state = States.idle;
            }

            // ANIMATION
            if (motion.y > 0)
            {
                if (direction == 1) animationPlayer.Play("WallSlideRight");
                else if (direction == -1) animationPlayer.Play("WallSlideLeft");
            }
            else
            {
                if (direction == 1) animationPlayer.Play("WallGrabRight");
                else if (direction == -1) animationPlayer.Play("WallGrabLeft");
            }
        }
        else if (state == States.crouch) // CROUCH
        {
            // BEHAVIOUR
            motion.x = 0;
            Gravity(delta);
            motion = MoveAndSlide(motion, upVector);

            // TRANSITION TRIGGERS
            if (Input.IsActionJustPressed("key_attack"))
            {
                state = States.attack;
            }
            else if (!IsOnFloor())
            {
                state = States.air;
            }
            else if (Input.IsActionJustReleased("key_down"))
            {
                if (x_input == 0)
                {
                    state = States.idle;
                }
                else
                {
                    state = States.run;
                }
            }

            // ANIMATION
            if (direction == 1) animationPlayer.Play("CrouchRight");
            else if (direction == -1) animationPlayer.Play("CrouchLeft");
        }
        else if (state == States.attack) // ATTACK
        {
            // BEHAVIOUR

            motion.x += x_input * ACCELERATION * delta;
            motion.x = Mathf.Clamp(motion.x, -MAX_SPEED, MAX_SPEED);
            if (x_input == 0) Friction(FRICTION);

            if (Input.IsActionJustReleased("key_jump"))
            {
                longJump = false;
                wallJumpTimer = 0;
                if (motion.y < 0)
                {
                    motion.y = 0.3f * motion.y;
                }
            }

            Gravity(delta);
            motion = MoveAndSlide(motion, upVector);

            if (IsOnFloor()) dashReloaded = true;
            if (Input.IsActionJustPressed("key_attack")) attackPressed = true;

            // TRANSITION TRIGGERS

            if ((Input.IsActionJustPressed("key_jump") || jumpBuffTimer > 0) && IsOnFloor())
            {
                motion.y = -JUMP_FORCE;
                PlaySound(asJump);
                jumpBuffTimer = 0;
                if (Input.IsActionPressed("key_jump")) longJump = true;
                state = States.air;
            }
            else if (dashReloadTimer == 0 && (Input.IsActionJustPressed("key_dash") || dashBuffTimer > 0) && canDash)
            {
                dashTimer = DASH_TIME;
                dashBuffTimer = 0;
                dashKeyWaitTimer = DASH_KEY_WAIT_TIME;
                PlaySound(asDash);
                camera2D.Call("StartShake", 0.3f, 0.01f, 1f, 1);
                state = States.dash;
            }
            else if (NextToLeftWall() && Input.IsActionPressed("key_left") && !IsOnFloor() && canWallJump)
            {
                onLeftWall = true;
                state = States.wallSlide;
            }
            else if (NextToRightWall() && Input.IsActionPressed("key_right") && !IsOnFloor() && canWallJump)
            {
                onLeftWall = false;
                state = States.wallSlide;
            }
            else if (animationPlayer.CurrentAnimationPosition > (animationPlayer.CurrentAnimationLength * 2 / 3) && attackPressed)
            {
                // proceeds with second attack animation if true
                attackPressed = false;
                attackIndex++;
                if (x_input != 0) direction = (int)x_input;
            }
            else if (!animationPlayer.IsPlaying()) // end of attack animation
            {
                attackIndex = 0;
                if (IsOnFloor())
                {
                    if (Input.IsActionPressed("key_down"))
                    {
                        state = States.crouch;
                    }
                    else if (x_input == 0)
                    {
                        state = States.idle;
                    }
                    else
                    {
                        state = States.run;
                    }
                }
                else
                {
                    dashReloaded = false;
                    state = States.air;
                }
            }

            // ANIMATION
            if (attackIndex >= 2) attackIndex = 0;
            if (animationPlayer.IsPlaying())
            {
                if (attackIndex == 0)
                {
                    if (direction == 1) animationPlayer.Play("AttackRight2");
                    else if (direction == -1) animationPlayer.Play("AttackLeft2");
                }
                else if (attackIndex == 1)
                {
                    if (direction == 1) animationPlayer.Play("AttackRight1");
                    else if (direction == -1) animationPlayer.Play("AttackLeft1");
                }
            }
        }
        else if (state == States.hurt) // HURT
        {
            // BEHAVIOUR
            motion.x = -direction * MAX_SPEED * 0.5f;
            Gravity(delta);
            motion = MoveAndSlide(motion, upVector);

            // TRANSITION TRIGGERS
            if (!animationPlayer.IsPlaying())
            {
                state = States.air;
            }
        }
        else if (state == States.death) // DEATH
        {
            // BEHAVIOUR
            motion.x = 0;
            Gravity(delta);
            motion = MoveAndSlide(motion, upVector);
        }
        else if (state == States.prepared) // PREPARED
        {
            // ANIMATION
            if (!IsOnFloor())
            {
                Gravity(delta);
                motion.x = 0;
                motion = MoveAndSlide(motion, upVector);

                animationPlayer.Play("AirFallRight");
            }
            else
            {
                if (!wasOnFloor)
                {
                    if (currentLevel == 1 && GlobalPosition.x < 100)
                    {
                        PlaySound(asImpactPrepared);
                    }
                    MakeDustParticle();
                    wasOnFloor = true;
                }
                animationPlayer.Play("CrouchRight");
                bool leftKeyPressed = Input.IsActionJustPressed("key_left");
                bool rightKeyPressed = Input.IsActionJustPressed("key_right");
                bool jumpKeyPressed = Input.IsActionJustPressed("key_jump");
                bool attackKeyPressed = Input.IsActionJustPressed("key_attack");
                if (leftKeyPressed || rightKeyPressed)
                {
                    state = States.run;
                    dashReloaded = true;
                    this.EmitSignal(nameof(fromPreparedState));
                }
                else if (jumpKeyPressed)
                {
                    MakeDustParticle();
                    motion.y = -JUMP_FORCE;
                    PlaySound(asJump);
                    jumpBuffTimer = 0;
                    if (Input.IsActionPressed("key_jump")) longJump = true;
                    state = States.air;
                    dashReloaded = true;
                    this.EmitSignal(nameof(fromPreparedState));
                }
                else if (attackKeyPressed)
                {
                    state = States.attack;
                    dashReloaded = true;
                    this.EmitSignal(nameof(fromPreparedState));
                }
            }
        }
        else if (state == States.noInput) // NO INPUT
        {
            Gravity(delta);
            Friction(0.1f);
            motion = MoveAndSlide(motion, upVector);

            // ANIMATION
            if (spikesTouched)
            {
                if (direction == 1) animationPlayer.Play("DeathRight");
                else if (direction == -1) animationPlayer.Play("DeathLeft");
            }
            else if (IsOnFloor())
            {
                animationPlayer.Play("IdleAlertRight");
            }
            else if (motion.y < 0)
            {
                animationPlayer.Play("AirUpRight");
            }
            else
            {
                animationPlayer.Play("AirFallRight");
            }
        }
    }

    public void Gravity(float delta)
    {
        motion.y = Mathf.Min(motion.y + GRAVITY * delta, MAX_FALL_SPEED);
    }

    public void Friction(float friction)
    {
        motion.x = Mathf.Lerp(motion.x, 0, friction);
    }

    public bool NextToWall()
    {
        return (NextToRightWall() || NextToLeftWall());
    }

    public bool NextToLeftWall()
    {
        return leftWallRayCast1.IsColliding() || leftWallRayCast2.IsColliding();
    }

    public bool NextToRightWall()
    {
        return rightWallRayCast1.IsColliding() || rightWallRayCast2.IsColliding();
    }

    public int GetLevelNumber()
    {
        string LevelName = GetTree().CurrentScene.Name;
        string LevelNumber = string.Empty;

        // for numbers with more than one digit
        var matches = Regex.Matches(LevelName, @"\d+");
        foreach (var match in matches)
        {
            // string addition
            LevelNumber += match;
        }

        return int.Parse(LevelNumber);
    }

    public int UpdateTimer(int timer)
    {
        if (timer > 0)
        {
            timer--;
            return timer;
        }
        return 0;
    }

    public int UpdateDirection(float input)
    {
        if (input != 0) return (int)input;
        else return direction;
    }

    public void MakeDustParticle()
    {
        Particles2D dust = (Particles2D)dustParticles.Instance();
        dust.GlobalPosition = new Vector2(GlobalPosition.x, GlobalPosition.y + 13);
        dust.Emitting = true;
        if (currentLevel <= 11)
        {
            dust.ProcessMaterial.Set("color", blueParticle);
        }
        else if (currentLevel <= 18)
        {
            dust.ProcessMaterial.Set("color", goldParticle);
        }
        else
        {
            dust.ProcessMaterial.Set("color", redParticle);
        }
        GetParent().AddChild(dust);
    }

    public void PlaySound(AudioStreamSample sample)
    {
        AudioStreamPlayer asp = (AudioStreamPlayer)audioStreamPlayer.Instance();
        asp.Stream = sample;
        playerVariables.AddChild(asp);
    }

    public void SetPlayerVariablesAfterDeath()
    {
        playerVariables.Set("Health", 5);
        playerVariables.Set("Souls", 0);
        playerVariables.Set("Level", playerVariables.Get("CheckpointLevel"));
        playerVariables.Set("Kills", playerVariables.Get("CheckpointKills"));
        playerVariables.Set("Berries", playerVariables.Get("CheckpointBerries"));
        playerVariables.Set("SwordDamage", playerVariables.Get("CheckpointSwordDamage"));

        if ((int)playerVariables.Get("CheckpointLevel") != 1)
        {
            playerVariables.Set("loaded", true);
        }
    }

    public void onHurtboxAreaEntered(Area2D area)
    {
        int health = (int)stats.Get("Health");
        bool lastLevelRocks = (area.Name == "RockHitbox" && currentLevel == LAST_LEVEL);

        if ((int)playerVariables.Get("Mode") != 1 && !lastLevelRocks)
        {
            health--;
        }

        if (health > 0)
        {
            camera2D.Call("StartShake", 0.3f, 0.01f, 1f, 1);
            camera2D.Call("StartFlash", 0.2f, 0.8f);
            PlaySound(asHurt);

            if (!lastLevelRocks)
            {
                hurtbox.Call("startInvincibility", INVINCIBILITY_TIME);
                hitAnimationPlayer.Play("Start");
            }

            if (area.Name == "SpikesHitbox" || area.Name == "LavaArea")
            {
                motion.x = 0;
                if (motion.y < 0) motion.y = 0;
                state = States.noInput;
                spikesTouched = true;
                spikeDeathTimer.Start(0.5f);

                if (direction == -1) animationPlayer.Play("DeathLeft");
                else if (direction == 1) animationPlayer.Play("DeathRight");

                animationPlayer.Seek(0.4f, false); // plays only second half of animation

                return;
            }

            if (health == 1)
            {
                worldAnimationPlayer.Play("BlurTransition");
            }

            motion.y = -JUMP_FORCE * 0.5f;
            state = States.hurt;

            if (direction == -1) animationPlayer.Play("HurtLeft");
            else if (direction == 1) animationPlayer.Play("HurtRight");
        }

        stats.Set("Health", health);
    }

    public void onSpikeDeathTimerTimeout()
    {
        if (GetLevelNumber() == LAST_LEVEL)
        {
            this.EmitSignal(nameof(lastLevelResetSignal));
        }
        respawnCollisionShape.SetDeferred("disabled", false);
        GlobalPosition = respawnPosition.GlobalPosition;
        state = States.prepared;
    }

    public void onDeathTransitionTimerTimeout()
    {
        camera2D.Call("FadeIn");
        transitionTimer.Start(0.5f);
        playerVariables.Set("makeTransition", true);
    }

    public void onTransitionTimerTimeout()
    {
        if ((int)playerVariables.Get("Mode") != 2)
        {
            GetTree().ChangeScene("res://Levels/Level_" + playerVariables.Get("CheckpointLevel") + ".tscn");
        }
        else
        {
            PlaySound(asLost);
            GetTree().ChangeScene("res://Menu/MainMenu.tscn");
            playerVariables.Call("DeleteData");
        }
    }

    // respawn after touching the spikes
    public void onRespawnAreaBodyEntered(KinematicBody2D body)
    {
        respawnCollisionShape.SetDeferred("disabled", true);

        hurtbox.Call("startInvincibility", INVINCIBILITY_TIME);
        hitAnimationPlayer.Play("Start");

        if ((int)playerVariables.Get("Mode") != 1 && GetLevelNumber() != LAST_LEVEL)
        {
            int health = (int)stats.Get("Health") - 1;
            stats.Set("Health", health);
        }

    }

    public void onInvisibilityTimerTimeout()
    {
        hitAnimationPlayer.Play("Stop");
    }

    public void onSwordHitboxBodyEntered(KinematicBody2D body)
    {
        camera2D.Call("StartShake", 0.05f, 0.01f, 1, 1);
    }

    public void onGhostTimerTimeout()
    {
        if (state == States.dash)
        {
            Sprite playerGhost = (Sprite)dashGhost.Instance();
            GetParent().AddChild(playerGhost);
            playerGhost.GlobalPosition = playerSprite.GlobalPosition;
            playerGhost.Texture = playerSprite.Texture;
            playerGhost.Vframes = playerSprite.Vframes;
            playerGhost.Hframes = playerSprite.Hframes;
            playerGhost.Frame = playerSprite.Frame;
            playerGhost.FlipH = playerSprite.FlipH;
            playerGhost.Offset = playerSprite.Offset;

        }
    }

    public void onNoHealthSignal()
    {
        worldEnvironment.Environment.AdjustmentSaturation = 0.5f;
        worldEnvironment.Environment.DofBlurNearDistance = 10;

        if (direction == -1) animationPlayer.Play("DeathLeft");
        else if (direction == 1) animationPlayer.Play("DeathRight");

        PlaySound(asDeath);

        hurtboxCollision.SetDeferred("disabled", true);
        hitboxCollision.SetDeferred("disabled", true);

        state = States.death;
        deathTransitionTimer.Start(1);

        SetPlayerVariablesAfterDeath();
    }
}


