using Godot;
using System;

public class Skeleton : KinematicBody2D
{
    // nodes
    Node stats;
    PlayerVariables playerVariables;
    AnimationPlayer animationPlayer;
    AnimationPlayer hitAnimationPlayer;
    RayCast2D leftRayCast;
    RayCast2D rightRayCast;
    RayCast2D leftPlayerDetect1;
    RayCast2D leftPlayerDetect2;
    RayCast2D rightPlayerDetect1;
    RayCast2D rightPlayerDetect2;
    KinematicBody2D player;
    Area2D swordHitbox;
    CollisionShape2D hurtboxCollision;
    CollisionShape2D hitboxCollision;
    Timer idleTimer;
    PackedScene audioStreamPlayer;
    AudioStreamSample asHit;
    // constants
    const int ACCELERATION = 1000;
    const int KNOCKBACK = 500;
    const int MAX_FALL_SPEED = 280;
    const int GRAVITY = 600;
    const float FRICTION = 0.3f;
    const int SOUL = 25;
    // vectors
    Vector2 upVector = new Vector2(0, -1);
    Vector2 motion = new Vector2(0, 0);
    // variables
    int playerDirection;
    int damage;
    bool dead = false;

    int health = 3;
    int direction = -1;

    //===========================

    enum States
    {
        idle,
        walk,
        attack,
        hurt,
        death,
        corpse,
        reborn
    }
    States state;


    public override void _Ready()
    {
        animationPlayer = GetNode<AnimationPlayer>("AnimationPlayer");
        hitAnimationPlayer = GetNode<AnimationPlayer>("HitAnimationPlayer");
        leftRayCast = GetNode<RayCast2D>("LeftRayCast");
        rightRayCast = GetNode<RayCast2D>("RightRayCast");
        leftPlayerDetect1 = GetNode<RayCast2D>("LeftPlayerDetect1");
        leftPlayerDetect2 = GetNode<RayCast2D>("LeftPlayerDetect2");
        rightPlayerDetect1 = GetNode<RayCast2D>("RightPlayerDetect1");
        rightPlayerDetect2 = GetNode<RayCast2D>("RightPlayerDetect2");
        idleTimer = GetNode<Timer>("IdleTimer");

        player = GetParent().GetNode<KinematicBody2D>("../Player");
        stats = player.GetNode<Node>("Stats");
        swordHitbox = GetParent().GetNode<Area2D>("../Player/HitboxPivot/SwordHitbox");
        hurtboxCollision = GetNode<CollisionShape2D>("Hurtbox/CollisionShape2D");
        hitboxCollision = GetNode<CollisionShape2D>("Hitbox/CollisionShape2D");
        playerVariables = GetNode<PlayerVariables>("/root/PlayerVariables");

        audioStreamPlayer = (PackedScene)ResourceLoader.Load("res://Sound/AudioStreamPlayer.tscn");
        asHit = (AudioStreamSample)ResourceLoader.Load("res://Sound/e_hit.wav");

        state = States.walk;
    }

    public override void _PhysicsProcess(float delta)
    {
        if (state == States.idle) // IDLE
        {
            // BEHAVIOUR
            motion.x = 0;
            Gravity(delta);
            motion = MoveAndSlide(motion, upVector);

            //TRANSITIONS TRIGGERS
            if ((direction == -1 && IsPlayerLeft()) || (direction == 1 && IsPlayerRight()))
            {
                idleTimer.Stop();
                state = States.attack;
            }

            // ANIMATION
            if (direction == -1) animationPlayer.Play("IdleLeft");
            else if (direction == 1) animationPlayer.Play("IdleRight");
        }
        else if (state == States.walk) // WALK 
        {
            // BEHAVIOUR
            if (IsOnWall()) direction = -direction;

            motion.x += direction * ACCELERATION * delta;
            Friction(FRICTION);
            Gravity(delta);
            motion = MoveAndSlide(motion, upVector);

            // TRANSITIONS TRIGGERS
            if ((!leftRayCast.IsColliding() && direction == -1) || (!rightRayCast.IsColliding() && direction == 1))
            {
                state = States.idle;
                idleTimer.Start(1);
            }
            else if ((direction == -1 && IsPlayerLeft()) || (direction == 1 && IsPlayerRight()))
            {
                state = States.attack;
            }

            // ANIMATION
            if (direction == -1) animationPlayer.Play("WalkLeft");
            else if (direction == 1) animationPlayer.Play("WalkRight");
        }
        else if (state == States.attack) // ATTACK 
        {
            // BEHAVIOUR
            Friction(FRICTION);
            Gravity(delta);
            motion = MoveAndSlide(motion, upVector);

            // ANIMATION
            if (direction == -1) animationPlayer.Play("AttackLeft");
            else if (direction == 1) animationPlayer.Play("AttackRight");
        }
        else if (state == States.hurt) // HURT 
        {
            // BEHAVIOUR
            Friction(FRICTION);
            Gravity(delta);
            motion = MoveAndSlide(motion, upVector);

            // ANIMATION
            if (direction == -1) animationPlayer.Play("HurtLeft");
            else if (direction == 1) animationPlayer.Play("HurtRight");
        }
        else if (state == States.death) // DEATH 
        {
            if (!dead)
            {
                dead = true;
                stats.Set("Souls", (int)stats.Get("Souls") + SOUL);
                stats.Set("Kills", (int)stats.Get("Kills") + 1);
            }
            // BEHAVIOUR
            motion.x = 0;
            Gravity(delta);
            motion = MoveAndSlide(motion, upVector);
        }
        else if (state == States.corpse) // CORPSE 
        {
            // TRANSITIONS TRIGGERS
            if (IsPlayerLeft() || IsPlayerRight())
            {
                state = States.reborn;

                if (IsPlayerLeft()) direction = -1;
                else direction = 1;
            }

            // ANIMATION
            animationPlayer.Play("CorpseLeft");
        }
        else if (state == States.reborn) // REBORN 
        {
            // ANIMATION
            if (direction == -1) animationPlayer.Play("RebornLeft");
            else if (direction == 1) animationPlayer.Play("RebornRight");
        }
    }

    public void Gravity(float delta)
    {
        motion.y = Mathf.Min(motion.y + GRAVITY * delta, MAX_FALL_SPEED);
    }

    public void Friction(float friction)
    {
        motion.x = Mathf.Lerp(motion.x, 0, friction);
    }

    public void PlaySound(AudioStreamSample sample)
    {
        AudioStreamPlayer asp = (AudioStreamPlayer)audioStreamPlayer.Instance();
        asp.Stream = sample;
        GetTree().CurrentScene.AddChild(asp);
    }

    public bool IsPlayerLeft()
    {
        return leftPlayerDetect1.IsColliding() || leftPlayerDetect2.IsColliding();
    }

    public bool IsPlayerRight()
    {
        return rightPlayerDetect1.IsColliding() || rightPlayerDetect2.IsColliding();
    }

    public void onHurtboxAreaEntered(Area2D area)
    {
        playerDirection = (int)player.Get("direction");
        damage = (int)swordHitbox.Get("damage");

        health -= damage;
        this.hitAnimationPlayer.Play("Start");
        PlaySound(asHit);

        if (health > 0)
        {
            if (state != States.attack)
            {
                motion.x = playerDirection * KNOCKBACK * 0.5f;
                direction = -playerDirection;
                state = States.hurt;
            }
        }
        else if (health <= 0)
        {
            if (direction == -1) animationPlayer.Play("DeathLeft");
            else if (direction == 1) animationPlayer.Play("DeathRight");

            hurtboxCollision.SetDeferred("disabled", true);
            hitboxCollision.SetDeferred("disabled", true);

            state = States.death;

            idleTimer.Stop();
        }
    }

    public void FromActionToIdle()
    {
        state = States.idle;
        idleTimer.Start(0.2f);
    }

    public void FromRebornToIdle()
    {
        if (IsPlayerLeft()) direction = -1;
        else if (IsPlayerRight()) direction = 1;

        state = States.idle;
        idleTimer.Start(0.2f);
    }

    public void onIdleTimerTimeout()
    {
        if ((direction == -1 && IsPlayerLeft()) || (direction == 1 && IsPlayerRight()))
        {
            state = States.attack;
        }
        else if (IsOnFloor())
        {
            state = States.walk;
            if ((!leftRayCast.IsColliding() && direction == -1) || (!rightRayCast.IsColliding() && direction == 1))
            {
                direction = -direction;
            }
        }
        else
        {
            state = States.idle;
            idleTimer.Start(2);
        }
    }
}

